/**
 * @file
 * Helper for entity reference proc field.
 */
(function($, Drupal, drupalSettings, once) {
  const ProcBehavior = {
    attach(context, settings) {
      once('proc-decrypt', 'html', context).forEach(initProcBehavior);
    },
  };

  /**
   * Initialize proc behavior.
   */
  function initProcBehavior() {
    // Get elements with 'proc' attribute set to 'true' that are not autocomplete.
    const procFieldElements = document.querySelectorAll(
      '[proc="true"]:not(.form-autocomplete)',
    );
    const procFields = Array.from(procFieldElements, extractFieldName);
    const encryptCheckboxes = document.querySelectorAll(
      '[id^="encrypt-checkbox-"]',
    );

    setTextSiblingsAttribute(procFields, procFieldElements, encryptCheckboxes);
    setInitialStateCheckboxes(procFieldElements, encryptCheckboxes);
    disableEncryptCheckboxIfEmpty(procFieldElements);
    switchSubmitOnInput(procFields, procFieldElements);
    switchSubmitOnCheckbox(encryptCheckboxes);
  }

  /**
   * Extract field name from field element.
   * @param fieldElement
   * @return fieldName
   */
  function extractFieldName(fieldElement) {
    const fieldName = fieldElement.attributes.name.value.split('[')[0];
    return fieldName;
  }

  /**
   * Set initial state of encrypt checkboxes.
   * @param procFieldElements
   * @param encryptCheckboxes
   */
  function setInitialStateCheckboxes(procFieldElements, encryptCheckboxes) {
    procFieldElements.forEach((fieldElement, index) => {
      encryptCheckboxes[index].disabled = fieldElement.value === '';
    });
  }

  /**
   * Disable encrypt checkbox if and only if field is empty.
   * @param procFieldElements
   */
  function disableEncryptCheckboxIfEmpty(procFieldElements) {
    procFieldElements.forEach((fieldElement, index) => {
      fieldElement.addEventListener('input', inputChangeCheckbox);
    });
  }

  /**
   * Encrypt checkbox change event handler.
   * @param inputEvent
   */
  function inputChangeCheckbox(inputEvent) {
    const fieldName = extractFieldName(inputEvent.target);
    const encryptCheckbox = document.querySelector(
      `[id="encrypt-checkbox-${fieldName}"]`,
    );
    encryptCheckbox.disabled = inputEvent.target.value === '';
  }

  /**
   * Enable submit if and only if all proc text fields are empty or disabled.
   * @param procFields
   * @param procFieldElements
   */
  function switchSubmitOnInput(procFields, procFieldElements) {
    // Process when there is a submit element specified in drupalSettings.
    if (drupalSettings.proc && drupalSettings.proc.submit_element_id) {
      procFields.forEach((fieldName, index) => {
        procFieldElements[index].setAttribute(
          'data-proc-text-siblings',
          procFields.join(),
        );
        procFieldElements[index].addEventListener('input', inputChangeSubmit);
      });
    }
  }

  /**
   * Set data-proc-text-siblings attributes.
   * @param procFields
   * @param procFieldElements
   * @param encryptCheckboxes
   */
  function setTextSiblingsAttribute(
    procFields,
    procFieldElements,
    encryptCheckboxes,
  ) {
    const procFieldsString = procFields.join();
    procFields.forEach((fieldName, index) => {
      procFieldElements[index].setAttribute(
        'data-proc-text-siblings',
        procFieldsString,
      );
      encryptCheckboxes[index].setAttribute(
        'data-proc-text-siblings',
        procFieldsString,
      );
    });
  }

  /**
   * Proc text field event change handler.
   * @param inputEvent
   */
  function inputChangeSubmit(inputEvent) {
    const submitElements = getSubmitElements();
    const siblings = getSiblings(inputEvent);

    const enable = siblings.every(sibling => {
      const siblingField = document.querySelector(`[name^="${sibling}"]`);
      return siblingField.value === '' || siblingField.style.display === 'none';
    });
    switchSubmit(submitElements, enable);
  }

  /**
   * Checkbox change event handler for changing submit.
   * @param checkBoxEvent
   */
  function checkboxChangeSubmit(checkBoxEvent) {
    const submitElements = getSubmitElements();
    const siblings = getSiblings(checkBoxEvent);

    const enable = siblings.every(sibling => {
      const siblingCheckbox = document.querySelector(
        `[id="encrypt-checkbox-${sibling}"]`,
      );
      return siblingCheckbox.checked || siblingCheckbox.disabled;
    });
    switchSubmit(submitElements, enable);
  }

  /**
   * Enable submit if and only if all checkboxes are checked.
   * @param encryptCheckboxes
   */
  function switchSubmitOnCheckbox(encryptCheckboxes) {
    // Process when there is a submit element specified in drupalSettings.
    if (drupalSettings.proc && drupalSettings.proc.submit_element_id) {
      encryptCheckboxes.forEach((encryptCheckbox, index) => {
        encryptCheckbox.addEventListener('input', checkboxChangeSubmit);
      });
    }
  }

  /**
   * Switch submit button.
   * @param submitElements
   * @param enable
   */
  function switchSubmit(submitElements, enable) {
    submitElements.forEach(submitElement => {
      // Switch disabled attribute only if necessary and when the submit element exists:
      if (submitElement && submitElement.disabled !== !enable) {
        submitElement.disabled = !enable;
      }
    });
  }

  /**
   * Get submit elements.
   * @return submitElements
   */
  function getSubmitElements() {
    const submitIds = drupalSettings.proc.submit_element_id
      .replace(/ /g, '')
      .split(',');
    return submitIds.map(submitId => document.getElementById(submitId));
  }

  /**
   * Get siblings.
   * @param event
   * @return siblings
   */
  function getSiblings(event) {
    const siblings = event.target
      .getAttribute('data-proc-text-siblings')
      .split(',');
    return siblings;
  }

  Drupal.behaviors.ProcBehavior = ProcBehavior;
})(jQuery, Drupal, drupalSettings, once);
