<?php

namespace Drupal\proc\Form;

use Drupal\Component\Utility\Crypt;
use Drupal\Component\Utility\UrlHelper;
use Drupal\Core\Ajax\AjaxResponse;
use Drupal\Core\Ajax\HtmlCommand;
use Drupal\Core\Ajax\InvokeCommand;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\File\FileSystem;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Link;
use Drupal\Core\Path\CurrentPathStack;
use Drupal\Core\Render\Markup;
use Drupal\Core\Session\AccountProxy;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\Core\Url;
use Drupal\file\Entity\File;
use Drupal\file\FileRepository;
use Drupal\file\FileUsage\DatabaseFileUsageBackend;
use Drupal\proc\Entity\Proc;
use Drupal\proc\ProcInterface;
use Drupal\proc\Traits\ProcMetadataMessageTrait;
use Drupal\proc\Traits\ProcRecipientTrait;
use Drupal\user\UserStorageInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\Request;

/**
 * Encrypt content.
 */
class ProcEncryptForm extends ProcOpFormBase {

  use ProcRecipientTrait, ProcMetadataMessageTrait;

  /**
   * The file system.
   *
   * @var \Drupal\Core\File\FileSystem
   */
  protected FileSystem $fileSystem;

  /**
   * The file repository.
   *
   * @var \Drupal\file\Entity\FileRepository
   */
  protected FileRepository $fileRepository;

  /**
   * The current user.
   *
   * @var \Drupal\Core\Session\AccountProxy
   */
  protected AccountProxy $currentUser;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected EntityTypeManagerInterface $entityTypeManager;

  /**
   * The user storage.
   *
   * @var \Drupal\user\UserStorageInterface
   */
  protected UserStorageInterface $userStorage;

  /**
   * The configuration factory.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * The current request.
   *
   * @var \Symfony\Component\HttpFoundation\Request
   */
  protected Request $request;

  /**
   * The label of the proc entity.
   *
   * @var string
   */
  protected String $label;

  /**
   * The current path.
   *
   * @var \Drupal\Core\Path\CurrentPathStack
   */
  protected $currentPath;

  /**
   * The module handler service.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected ModuleHandlerInterface $moduleHandler;

  /**
   * The file usage service.
   *
   * @var \Drupal\file\FileUsage\DatabaseFileUsageBackend
   */
  protected DatabaseFileUsageBackend $fileUsage;

  /**
   * ProcEncryptForm form.
   *
   * @param \Drupal\Core\File\FileSystem $file_system
   *   The file system.
   * @param \Drupal\file\Entity\FileRepository $file_repository
   *   The file repository.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\Core\Session\AccountProxy $current_user
   *   The user account.
   * @param \Drupal\user\UserStorageInterface $user_storage
   *   The user storage.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The config factory.
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   The current request.
   * @param \Drupal\Core\Extension\ModuleHandlerInterface $module_handler
   *   The module handler.
   * @param \Drupal\file\FileUsage\DatabaseFileUsageBackend $file_usage
   *   The file usage service.
   */
  public function __construct(
    FileSystem $file_system,
    FileRepository $file_repository,
    EntityTypeManagerInterface $entity_type_manager,
    AccountProxy $current_user,
    UserStorageInterface $user_storage,
    ConfigFactoryInterface $config_factory,
    Request $request,
    CurrentPathStack $current_path,
    ModuleHandlerInterface $module_handler,
    DatabaseFileUsageBackend $file_usage,
  ) {
    $this->fileSystem = $file_system;
    $this->fileRepository = $file_repository;
    $this->entityTypeManager = $entity_type_manager;
    $this->currentUser = $current_user;
    $this->userStorage = $user_storage;
    $this->configFactory = $config_factory;
    $this->request = $request;
    $this->currentPath = $current_path;
    $this->moduleHandler = $module_handler;
    $this->fileUsage = $file_usage;
  }

  /**
   * {@inheritdoc}
   *
   * @param \Symfony\Component\DependencyInjection\ContainerInterface $container
   *   The Drupal service container.
   *
   * @return static
   */
  public static function create(ContainerInterface $container): static {
    return new static(
      $container->get('file_system'),
      $container->get('file.repository'),
      $container->get('entity_type.manager'),
      $container->get('current_user'),
      $container->get('entity_type.manager')->getStorage('user'),
      $container->get('config.factory'),
      $container->get('request_stack')->getCurrentRequest(),
      $container->get('path.current'),
      $container->get('module_handler'),
      $container->get('file.usage')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId(): string {
    return 'proc_encrypt_form';
  }

  /**
   * Dynamic title for Encrypt form.
   */
  public function getTitle(): TranslatableMarkup {
    $title = $this->t('Encrypt and upload');

    if (!$this->isFileInputMode()) {
      $title = $this->t('Encrypt');
    }

    return $title;
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state): array {
    // Define mappings for field descriptions and button values.
    $field_description_mapping = [
      $this->t('Select a file for encryption.'),
      $this->t('Compose a story for encryption.'),
      $this->t('Write a text for encryption.'),
      $this->t('Choose a date for encryption.'),
    ];

    $dialog_button_value_mapping = [
      $this->t('Encrypt and upload'),
      $this->t('Encrypt'),
      $this->t('Encrypt'),
      $this->t('Encrypt'),
    ];

    // Add hidden fields.
    foreach (ProcInterface::HIDDEN_FIELDS as $hidden_field) {
      $form[$hidden_field] = ['#type' => 'hidden'];
    }

    // Retrieve query parameters.
    $query = $this->getRequest()->query->all();

    // Set host_page_url if available in the query.
    $form['host_page_url']['#value'] = $query['host_page_url'] ?? NULL;

    // Set field input mode.
    $field_input_mode = $query['field-input-mode'] ?? 0;

    // Add field based on input mode.
    $form[ProcInterface::FIELD_KEYS[$field_input_mode]] = [
      '#type' => ProcInterface::FIELD_TYPES[$field_input_mode],
      '#description' => $field_description_mapping[$field_input_mode],
    ];

    // Set Drupal settings for standalone mode.
    $proc_standalone_mode = $query['proc_standalone_mode'] ?? FALSE;
    $form['#attached']['drupalSettings']['proc']['proc_standalone_mode'] = $proc_standalone_mode;

    // Add submit button with Ajax.
    $form['submit-proc'] = [
      '#type' => 'submit',
      '#value' => $dialog_button_value_mapping[$field_input_mode],
      '#ajax' => [
        'callback' => '::submitFormAjax',
        'event' => 'click',
      ],
    ];

    // Adjust form elements if not in standalone mode.
    if ($proc_standalone_mode != 'FALSE') {
      $form['submit-proc']['#ajax'] = NULL;
      $form['actions']['submit-proc'] = $form['submit-proc'];
      unset($form['submit-proc']);
    }

    // Retrieve selected user IDs from the current path.
    $current_path = $this->currentPath->getPath();
    $selected_user_ids = explode(',', explode('/', $current_path)[3]);

    // Retrieve creation dates of the newest encryption keys for each user.
    $created = [];
    foreach ($selected_user_ids as $user_id) {
      if (is_numeric($user_id)) {
        $proc_id = $this->getNewestProcId($user_id);
        if ($proc_id) {
          $proc = Proc::load($proc_id);
          if ($proc) {
            $created[$user_id] = $proc->get('created')->value;
          }
        }
      }
    }

    // Set additional Drupal settings.
    $settings = [
      'proc_labels' => _proc_js_labels(),
      'proc_data' => _proc_common_get_decryption_form_data(NULL),
      'proc_file_entity_max_filesize' => $this->configFactory->get('proc.settings')->get('proc-file-entity-max-filesize') ?? 10000000,
      'proc_post_max_size_bytes' => _proc_get_post_max_size_bytes(),
      'proc_recipients_pubkeys_changed' => json_encode($created),
      'basePath' => base_path(),
      'proc_form_triggering_field' => $query['proc_parent_id'] ?? '',
      'proc_field_delta' => $query['proc_field_delta'] ?? '',
      'proc-field-input-mode' => $field_input_mode ?? '0',
      'proc_field_name' => $query['proc_field_name'] ?? '',
      'host_page_url' => $query['host_page_url'] ?? '',
    ];
    $settings += $form['#attached']['drupalSettings']['proc'];

    // Attach libraries and settings.
    $form['#attached'] = [
      'library' => [
        0 => 'proc/openpgpjs',
        1 => ProcInterface::PROC_ENCRYPTION_LIBRARIES[$field_input_mode],
      ],
      'drupalSettings' => $settings,
    ];

    // Set form action.
    $current_path = mb_substr($current_path, 1);
    $form_state->set('storage', $selected_user_ids);
    $form['#action'] = base_path() . $current_path;

    // Adjust form action based on destination.
    $destination = isset($query['destination']) && $proc_standalone_mode !== 'FALSE' ? $query['destination'] : FALSE;
    if ($destination) {
      $form['#action'] = base_path() . $destination;
    }

    return $form;
  }

  /**
   * Get the newest Proc ID for a given user ID.
   */
  private function getNewestProcId($user_id) {
    $query = $this->entityTypeManager->getStorage('proc')->getQuery();
    $query->accessCheck(TRUE);
    $query->condition('user_id', $user_id);
    $query->condition('type', 'cipher', '!=');
    $query->sort('id', 'DESC');
    $query->range(0, 1);
    $proc_ids = $query->execute();
    return !empty($proc_ids) ? reset($proc_ids) : NULL;
  }

  /**
   * Submit handler for the ajax submit.
   */
  public function submitFormAjax(array &$form, FormStateInterface $form_state): AjaxResponse {
    $field_name = $form_state->getCompleteForm()['#attached']['drupalSettings']['proc_field_name'];
    // Close the dialog:
    $response = new AjaxResponse();
    // Selector for the dialog close button.
    $selector = '.ui-dialog-titlebar-close.class-' . $field_name;
    $method = 'click';
    $arguments = [];
    $response->addCommand(new InvokeCommand($selector, $method, $arguments));

    $file_name = (isset($form['proc-file']['#value']))
      ? $form['proc-file']['#value']["\0Symfony\\Component\\HttpFoundation\\File\\UploadedFile\0originalName"]
      : '';

    $proc_id = FALSE;
    // Get the latest proc ID authored by current user:
    $current_user_id = $this->currentUser->id();

    // If this proc was edited instead of created, use the known proc ID:
    $current_url = $form['submit-proc']['#attached']['drupalSettings']['ajaxTrustedUrl'];
    $current_url_string = key($current_url);
    $parsed_url = UrlHelper::parse($current_url_string);
    $edited_proc_id = FALSE;
    if (isset($parsed_url['query']['field-default-proc-id'])) {
      $edited_proc_id = $parsed_url['query']['field-default-proc-id'];
    }
    $proc_entity = FALSE;
    if ($edited_proc_id) {
      $proc_entity = Proc::load($edited_proc_id);
    }

    if ($proc_entity) {
      if ($proc_entity->getOwnerId() == $this->currentUser->id()) {
        $proc_id = $edited_proc_id;
      }
      else {
        $recipients = $proc_entity->get('field_recipients_set')->getValue();
        foreach ($recipients as $recipient) {
          if ($recipient['target_id'] == $this->currentUser->id()) {
            $proc_id = $edited_proc_id;
            break;
          }
        }
      }
    }

    // If there is no default proc ID, get the latest proc ID authored by current user:
    if (!$proc_id && !$edited_proc_id) {
      $query = $this->entityTypeManager->getStorage('proc')->getQuery();
      $query->accessCheck(TRUE);
      $query->condition('user_id', $current_user_id);
      $query->sort('id', 'DESC');
      $query->range(0, 1);
      $entity_ids = $query->execute();
      $proc_id = array_shift($entity_ids);
    }

    $new_proc_entity = FALSE;
    if ($proc_id) {
      $new_proc_entity = Proc::load($proc_id);
      $proc_object_label = $new_proc_entity->get('label')->getValue()[0]['value'];
    }

    // If label was altered by hook proc label alter, use it:
    if (!empty($file_name) && $this->label != $file_name) {
      $file_name = $this->label;
    }
    $metadataSelector = FALSE;
    if (!empty($file_name) && $file_name == $proc_object_label) {
      // Set the file name in the form:
      $delta = $form_state->getCompleteForm()['#attached']['drupalSettings']['proc_field_delta'];
      $selector = 'input[name="' . $field_name . '[' . $delta . '][target_id]"]';
      $file_name .= ' (' . $proc_id . ')';
      // Replace _ by - in field naem:
      $metadataSelector = '#edit-' . str_replace('_', '-', $field_name) . '-' . $delta . '-metadata';
    }
    if (empty($file_name)) {
      $selector = '#' . $form_state->getCompleteForm()['#attached']['drupalSettings']['proc_form_triggering_field'];
      // Text field encryption mode uses a hidden textfield as input mode and
      // therefore accepts only the cipher text ID:
      $file_name .= $proc_id;
    }

    // If the file name is not numeric, it is a text field encryption mode:
    if (!is_numeric($file_name) && $metadataSelector) {
      if (!$proc_entity && $new_proc_entity) {
        $proc_entity = $new_proc_entity;
      }
      $encryptionDateTimeMessage = $this->getMetadataMessage($proc_entity);
      $response->addCommand(new HtmlCommand($metadataSelector, $encryptionDateTimeMessage));
    }

    $response->addCommand(new InvokeCommand($selector, 'val', [$file_name]));

    return $response;
  }

  /**
   * Implements form validation.
   *
   * @param array $form
   *   The render array of the currently built form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   Object describing the current state of the form.
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {
    // Validate format of cipher text by header string:
    $cipher_text = $form_state->getValue('cipher_text');
    $header = explode("\n", $cipher_text)[0];
    if (strpos($header, '-----BEGIN PGP MESSAGE-----') === FALSE) {
      $form_state->setErrorByName('cipher_text', $this->t('Invalid cipher text format.'));
    }
  }

  /**
   * Implements a form submit handler.
   *
   * @param array $form
   *   The render array of the currently built form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   Object describing the current state of the form.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public function submitForm(array &$form, FormStateInterface $form_state): void {
    $recipients_set_ids = $form_state->get('storage');
    $current_url = $this->request->headers->get('referer');
    $json_content = $form_state->getValue('cipher_text');
    $meta = $this->extractMetaFields($form_state);
    $field_default_proc_id = $this->getFieldDefaultProcId($current_url);
    $proc = $this->createOrUpdateProcEntity($field_default_proc_id);

    $user_evaluation = $this->getRecipientUsers($recipients_set_ids);
    $recipient_users = $user_evaluation['recipient_users'];
    $rejected_users = $user_evaluation['rejected_users'];

    $files = $this->saveJsonFiles($json_content);
    $file_id = $files['file_id'];
    $json_fids = $files['json_fids'];

    $cipher = '';
    if ($file_id) {
      $cipher = ['cipher_fid' => $file_id];
    }
    elseif (!empty($json_fids)) {
      $cipher = ['cipher_fid' => $json_fids];
    }

    $label = $meta['source_file_name'];
    // Implement hook proc label alter.
    $this->moduleHandler->alter('proc_label', $label, $meta, $form_state);

    $this->label = $label;

    $proc->set('armored', $cipher)
      ->set('meta', $meta)
      ->set('langcode', 'en')
      ->set('label', $label)
      ->set('type', 'cipher')
      ->set('field_recipients_set', $recipient_users)
      ->save();

    $proc_id = $proc->id();

    if (is_numeric($proc_id)) {
      $this->handleFileUsage($file_id, $proc_id, $json_fids);

      $link = $this->generateLink($proc_id);
      $this->logEncryptionCompletion($link);
      $this->handleFeedback($proc_id, $rejected_users);
    }
    else {
      $this->messenger()->addError($this->t('Error'));
    }
  }

  /**
   * Handles file usage.
   */
  private function handleFileUsage($file_id, $proc_id, $json_fids): void {
    if ($file_id) {
      $this->addFileUsage($file_id, $proc_id);
    }

    if (!empty($json_fids)) {
      foreach ($json_fids as $json_fid) {
        $this->addFileUsage($json_fid, $proc_id);
      }
    }
  }

  /**
   * Adds file usage.
   */
  private function addFileUsage($file_id, $proc_id): void {
    $file = File::load($file_id);
    $file_usage = $this->fileUsage;
    $file_usage->add($file, 'proc', 'file', $proc_id);
  }

  /**
   * Generates the decryption link.
   */
  private function generateLink($proc_id): string {
    global $base_url;
    $url = Url::fromUri('internal:/proc/' . $proc_id);
    return Link::fromTextAndUrl($base_url . '/proc/' . $proc_id, $url)
      ->toString()
      ->getGeneratedLink();
  }

  /**
   * Logs encryption completion.
   */
  private function logEncryptionCompletion($link): void {
    $this->getLogger('proc')->notice('Encryption is completed. Decryption link: %link', ['%link' => Markup::create($link)]);
  }

  /**
   * Handles feedback.
   */
  private function handleFeedback($proc_id, $rejected_users): void {
    $config = $this->configFactory->get('proc.settings');

    if (!$config->get('proc-suppress-encryption-success-msg')) {
      $link = $this->generateLink($proc_id);

      if (!$config->get('proc-suppress-decrypt-link')) {
        $this->messenger()->addMessage($this->t('Encryption is completed. Decryption link: %link', ['%link' => Markup::create($link)]));
      }
      else {
        $this->messenger()->addMessage($this->t('Encryption is completed.'));
      }
    }

    if (!empty($rejected_users)) {
      $rejected_user_ids_csv = implode(', ', $rejected_users);
      $this->logRecipientImpact($proc_id, $rejected_user_ids_csv);
      $this->messenger()->addWarning($this->t('PGP key(ies) or recipient(s) not found on encrypted content (id: @proc_id). Rejected user ID(s): @rejected_users.', [
        '@proc_id' => $proc_id,
        '@rejected_users' => $rejected_user_ids_csv,
      ]));
    }
  }

  /**
   * Logs recipient impact.
   */
  private function logRecipientImpact($proc_id, $rejected_user_ids_csv): void {
    $this->getLogger('proc')->warning('PGP key(ies) or recipient(s) not found on encrypted content (id: @proc_id). Rejected user ID(s): @rejected_users.', [
      '@proc_id' => $proc_id,
      '@rejected_users' => $rejected_user_ids_csv,
    ]);
  }

  /**
   * Extract meta fields from the form state.
   *
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   Object describing the current state of the form.
   *
   * @return array
   *   Array containing meta field values.
   */
  private function extractMetaFields(FormStateInterface $form_state): array {
    return [
      'source_file_name' => $form_state->getValue('source_file_name'),
      'source_file_size' => $form_state->getValue('source_file_size'),
      'source_file_type' => $form_state->getValue('source_file_type'),
      'source_file_last_change' => $form_state->getValue('source_file_last_change'),
      'browser_fingerprint' => $form_state->getValue('browser_fingerprint'),
      'generation_timestamp' => $form_state->getValue('generation_timestamp'),
      'generation_timespan' => $form_state->getValue('generation_timespan'),
      'host_page_url' => $form_state->getValue('host_page_url'),
      'signed' => $form_state->getValue('signed'),
    ];
  }

  /**
   * Get the default proc ID from the current URL.
   *
   * @param string $current_url
   *   The current URL.
   *
   * @return int|false
   *   The default proc ID or false if not found.
   */
  private function getFieldDefaultProcId(string $current_url): bool|int {
    $parsed_url = UrlHelper::parse($current_url);
    $query_string = $this->request->get('field-default-proc-id');
    if (isset($parsed_url['query']['field-default-proc-id'])) {
      return (int) $parsed_url['query']['field-default-proc-id'];
    }
    if (isset($query_string)) {
      return (int) $query_string;
    }
    return FALSE;
  }

  /**
   * Get recipient users based on the selected strategy.
   *
   * @param array $recipients_set_ids
   *   Recipients set IDs.
   *
   * @return array
   *   Array of recipient users and/or rejected users.
   */
  private function getRecipientUsers(array $recipients_set_ids): array {
    $recipient_users = [];
    $rejected_users = [];

    switch ($this->configFactory->get('proc.settings')->get('proc-auto-filter-recipients')) {
      case ProcInterface::STRATEGY_RESTRICTIVE:
        foreach ($recipients_set_ids as $recipient_id) {
          $recipient_users[] = ['target_id' => $recipient_id];
        }
        break;

      case ProcInterface::STRATEGY_PERMISSIVE:
        foreach ($recipients_set_ids as $recipient_id) {
          if ($this->userHasKeyring($recipient_id)) {
            $recipient_users[] = ['target_id' => $recipient_id];
          }
          else {
            $rejected_users[] = $recipient_id;
          }
        }
        break;
    }

    return ['recipient_users' => $recipient_users, 'rejected_users' => $rejected_users];
  }

  /**
   * Save JSON content as separate files if needed.
   *
   * @param string $json_content
   *   JSON content to be saved.
   *
   * @return array
   *   Array of file IDs.
   */
  private function saveJsonFiles(string $json_content): array {
    $config = $this->configFactory->get('proc.settings');
    $enable_stream_wrapper = $config->get('proc-enable-stream-wrapper');
    $stream_wrapper = $config->get('proc-stream-wrapper');

    $json_fids = [];
    $file_id = FALSE;

    if ($enable_stream_wrapper === 1 && !empty($stream_wrapper)) {

      $block_size = $config->get('proc-file-block-size');

      $blocks_split_enabled = $config->get('proc-enable-block-size');

      $json_dest = $stream_wrapper;
      $this->createDirectory($json_dest);

      if ($json_content) {
        if ($blocks_split_enabled && !empty($block_size)) {
          $blocks_texts = $this->splitContentIntoBlocks($json_content, $block_size);
          foreach ($blocks_texts as $block_text) {
            $json_fids[] = $this->writeJsonBlock($json_dest, $block_text);
          }
        }
        else {
          $file_id = $this->writeJsonBlock($json_dest, $json_content);
        }
      }
    }
    return ['json_fids' => $json_fids, 'file_id' => $file_id];
  }

  /**
   * Create directory if it does not exist.
   *
   * @param string $directory
   *   The directory path.
   */
  private function createDirectory(string $directory): void {
    if (!is_dir($directory)) {
      $this->fileSystem->mkdir($directory, NULL, TRUE);
    }
  }

  /**
   * Split content into blocks based on block size.
   *
   * @param string $json_content
   *   JSON content to be split.
   * @param int $block_size
   *   Size of each block.
   *
   * @return array
   *   Array of content blocks.
   */
  private function splitContentIntoBlocks(string $json_content, int $block_size): array {
    $lines = explode("\n", $json_content);
    $content_lines_number = count($lines);
    $lines_size_ratio = $content_lines_number / $block_size;
    $blocks = intval($lines_size_ratio);
    $remaining = $content_lines_number % $block_size;

    if ($remaining > 0) {
      $blocks++;
    }

    $blocks_lines = [];
    $content_line_index = 0;
    $blocks_index = 0;

    while ($blocks_index < $blocks) {
      $line_in_block_index = 0;
      while ($line_in_block_index < $block_size) {
        if (isset($lines[$content_line_index])) {
          $blocks_lines[$blocks_index][] = $lines[$content_line_index];
        }
        $content_line_index++;
        $line_in_block_index++;
      }
      $blocks_index++;
    }
    $blocks_texts = [];
    foreach ($blocks_lines as $block_index => $block_lines) {
      foreach ($block_lines as $block_line) {
        $blocks_texts[$block_index] = $blocks_texts[$block_index] . "\n" . $block_line;
      }
    }

    return $blocks_texts;
  }

  /**
   * Write JSON block to a file and return the file ID.
   *
   * @param string $json_dest
   *   Destination directory for JSON files.
   * @param string $block_text
   *   JSON block content.
   *
   * @return int
   *   File ID of the written JSON block.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  private function writeJsonBlock(string $json_dest, string $block_text): int {
    $fragment = Crypt::hashBase64(Crypt::randomBytesBase64(32));
    $json_filename = $fragment . '.json';
    $jsonFid = $this->fileRepository->writeData(
      $block_text,
      "$json_dest/$json_filename",
      FileSystemInterface::EXISTS_REPLACE
    );

    return $jsonFid->id();
  }

  /**
   * Validate Proc field input mode.
   */
  private function isFileInputMode(): bool {
    $input_mode = ($this->request->query->has('field-input-mode'))
      ? $this->request->query->get('field-input-mode')
      : NULL;
    if (!isset($input_mode)) {
      return FALSE;
    }

    return ((int) $input_mode === ProcInterface::INPUT_MODE_FILE);
  }

}
