<?php

namespace Drupal\proc\Form;

use Drupal\Component\Utility\UrlHelper;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Url;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\RedirectResponse;

/**
 * Generate PGP asymmetric keys.
 */
class ProcKeysGenerationForm extends ProcOpFormBase {

  /**
   * The configuration factory.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * ProcKeysGenerationForm form.
   *
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The config factory.
   */
  public function __construct(
    ConfigFactoryInterface $config_factory,
  ) {
    $this->configFactory = $config_factory;
  }

  /**
   * {@inheritdoc}
   *
   * @param \Symfony\Component\DependencyInjection\ContainerInterface $container
   *   The Drupal service container.
   *
   * @return static
   */
  public static function create(ContainerInterface $container): static {
    return new static(
      $container->get('config.factory'),
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId(): string {
    return 'proc_keys_generation_form';
  }

  /**
   * Build the simple form.
   *
   * A build form method constructs an array that defines how markup and
   * other form elements are included in an HTML form.
   *
   * @param array $form
   *   Default form array structure.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   Object containing current form state.
   *
   * @return array
   *   The render array defining the elements of the form.
   */
  public function buildForm(array $form, FormStateInterface $form_state): array {
    $proc_hidden_fields_key_generation = [
      'public_key',
      'encrypted_private_key',
      'generation_timestamp',
      'generation_timespan',
      'browser_fingerprint',
      'proc_email',
    ];

    foreach ($proc_hidden_fields_key_generation as $hidden_field) {
      $form[$hidden_field] = ['#type' => 'hidden'];
    }

    // Password Confirm.
    $form['password_confirm'] = [
      '#type' => 'password_confirm',
      '#required' => TRUE,
    ];

    // Group submit handlers in an actions element with a key of "actions" so
    // that it gets styled correctly, and so that other modules may add actions
    // to the form. This is not required, but is convention.
    $form['actions']['submit'] = [
      '#type' => 'submit',
      '#value' => $this->t('Submit'),
    ];

    $proc_data = _proc_common_get_decryption_form_data(NULL);

    $key_size = $this->configFactory->get('proc.settings')->get('proc-rsa-key-size');

    $form['#attached'] = [
      'library' => [
        0 => 'proc/openpgpjs',
        1 => 'proc/proc-generate-keys',
      ],
      'drupalSettings' => [
        'proc' => [
          'proc_labels' => _proc_js_labels(),
          'proc_data' => $proc_data,
          'proc_key_size' => $key_size,
        ],
      ],
    ];

    $form['#action'] = 'generate-keys';

    return $form;
  }

  /**
   * Implements form validation.
   *
   * The validateForm method is the default method called to validate input on
   * a form.
   *
   * @param array $form
   *   The render array of the currently built form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   Object describing the current state of the form.
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {
    // Check if the cipher text is empty.
    $public_key = $form_state->getValue('public_key');
    if (empty($public_key)) {
      // Set error for entire form:
      $form_state->setError($form, $this->t('Unknown error on creating asymmetric keys.'));

    }
    // Check if the generated keys look like PGP keys.
    if (strpos($public_key, '-----BEGIN PGP PUBLIC KEY BLOCK-----') !== 0) {
      $form_state->setError($form, $this->t('Invalid PGP public key format.'));
    }
  }

  /**
   * Implements a form submit handler.
   *
   * The submitForm method is the default method called for any submit elements.
   *
   * @param array $form
   *   The render array of the currently built form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   Object describing the current state of the form.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public function submitForm(array &$form, FormStateInterface $form_state): void {
    $destination = FALSE;
    $key_created = FALSE;
    $success_message = $this->t('Your key is saved.');

    $current_url = $this->getRequest()->getSchemeAndHttpHost() . $this->getRequest()->getRequestUri();

    $parse_result = UrlHelper::parse($current_url);
    if (isset($parse_result['query']['destination'])) {
      $destination = $parse_result['query']['destination'];
    }

    if (!empty($form_state->getValue('encrypted_private_key'))) {
      $keyring = [
        'privkey' => $form_state->getValue('encrypted_private_key'),
        'pubkey' => $form_state->getValue('public_key'),
      ];

      $config = $this->config('proc.settings');
      $key_size = $config->get('proc-rsa-key-size');

      $meta = [
        'generation_timestamp' => $form_state->getValue('generation_timestamp'),
        'generation_timespan' => $form_state->getValue('generation_timespan'),
        'browser_fingerprint' => $form_state->getValue('browser_fingerprint'),
        'proc_email' => $form_state->getValue('proc_email'),
        'key_size' => $key_size,
      ];

      $proc = $this->createOrUpdateProcEntity(NULL);

      $proc->set('armored', $keyring)
        ->set('meta', $meta)
        ->set('label', $meta['proc_email'])
        ->save();

      $key_created = TRUE;
      if (!$destination) {
        $this->messenger()->addMessage($success_message);
      }
    }
    else {
      $this->messenger()
        ->addMessage($this->t('Unknown error on the generation of the key. Please try again.'), 'error');
    }

    if ($destination && $key_created) {
      $url = Url::fromUri('internal:/' . $destination);
      $response = new RedirectResponse($url->toString());
      $response->send();
      $this->messenger()->addMessage($success_message);
    }
  }

}
