<?php

namespace Drupal\proc\Plugin\Field\FieldWidget;

use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Field\Plugin\Field\FieldWidget\EntityReferenceAutocompleteWidget;
use Drupal\Core\Form\FormStateInterface;
use Drupal\proc\Element\ProcField;
use Drupal\proc\Entity\Proc;
use Drupal\proc\Event\ProcElementEvent;
use Drupal\proc\Event\ProcEvents;
use Drupal\proc\ProcInterface;

/**
 * Defines the 'proc_entity_reference_widget' field widget.
 *
 * @FieldWidget(
 *   id = "proc_entity_reference_widget",
 *   module = "proc",
 *   label = @Translation("Proc Entity Reference Field Widget"),
 *   field_types = {
 *     "proc_entity_reference_field"
 *   }
 * )
 */
class ProcEntityReferenceWidget extends EntityReferenceAutocompleteWidget {

  /**
   * Recipients fetcher endpoint URL.
   *
   * @var string
   */
  private $recipientsFetcher;

  /**
   * Form element IDs to be used as parameters for the recipients endpoint.
   *
   * @var string
   */
  private $fetcherFilter;

  /**
   * Type of identifier to be used as parameters for the recipients endpoint.
   *
   * 0: ID
   * 1: Name.
   *
   * @var int
   */
  private $fetcherFilterType;

  /**
   * Comma separated list of user IDs to be set as recipients for a encryption.
   *
   * @var string
   */
  private $defaultRecipients;

  /**
   * Machine name of a field containing user IDs of recipients.
   *
   * @var int
   */
  private $recipientsTo;

  /**
   * Machine name of a field containing user IDs of CC recipients.
   *
   * @var int
   */
  private $recipientsCc;

  /**
   * Mode of operation.
   *
   * @var int
   */
  private $opMode;

  /**
   * Modes of data input.
   *
   * 0: Disabled
   * 1: Only encryption
   * 2: Encryption and signature.
   *
   * @var int
   */
  private $inputMode;

  /**
   * Modes of decryption.
   *
   * 0: New page
   * 1: Dialog.
   *
   * @var int
   */
  private $decryptionMode;

  /**
   * Password caching flag.
   *
   * 0: Disable password caching
   * 1: 'Show checkbox for allowing password caching
   * 2: 'Hide the checkbox and apply password caching by default.
   *
   * @var int
   */
  private $passCaching;

  /**
   * Trigger format.
   *
   * 0: Encrypt and Decrypt buttons
   * 1: Toggle checkbox (check for encryption, uncheck for decryption)
   *
   * @var int
   */
  private $triggerFormat;

  /**
   * Form submit element ID for the master field.
   *
   * @var string
   */
  private $masterFieldSubmitId;

  /**
   * Flag for triggering decryption of another encrypted field.
   *
   * @var int
   */
  private $triggerDecryption;

  /**
   * Return field operation mode options.
   */
  private function getFieldOperationModes() {
    return [
      ProcInterface::FIELD_OPS_DISABLED  => $this->t('Disabled'),
      ProcInterface::FIELD_OPS_ONLY_ENCRYPTION => $this->t('Only encryption'),
      ProcInterface::FIELD_OPS_ENCRYPTION_AND_SIGNATURE => $this->t('Encryption and signature'),
    ];
  }

  /**
   * Return field input mode options.
   */
  private function getInputModes() {
    return [
      ProcInterface::INPUT_MODE_FILE => $this->t('File'),
      ProcInterface::INPUT_MODE_TEXTAREA => $this->t('Text area'),
      ProcInterface::INPUT_MODE_TEXTFIELD => $this->t('Text field'),
      ProcInterface::INPUT_MODE_DATE => $this->t('Date'),
    ];
  }

  /**
   * Return decryption mode options.
   */
  private function getDecryptionModes() {
    return [
      ProcInterface::DECRYPT_MODE_NEW_PAGE => $this->t('New page'),
      ProcInterface::DECRYPT_MODE_DIALOG => $this->t('Dialog'),
    ];
  }

  /**
   * Return password caching options.
   */
  private function getPasswordCachingOptions() {
    return [
      ProcInterface::PASS_CACHE_OPS_DISABLED => $this->t('Disable password caching'),
      ProcInterface::PASS_CACHE_OPS_ALLOW => $this->t('Show checkbox for allowing password caching'),
      ProcInterface::PASS_CACHE_OPS_APPLY => $this->t('Hide the checkbox and apply password caching by default'),
    ];
  }

  /**
   * Return trigger format options.
   */
  private function getTriggerFormats() {
    return [
      ProcInterface::TRIGGER_FORMAT_BUTTONS => $this->t('Encrypt and Decrypt buttons'),
      ProcInterface::TRIGGER_FORMAT_CHECKBOX => $this->t('Toggle checkbox (check for encryption, uncheck for decryption)'),
    ];
  }

  /**
   * Return fetcher endpoint filter type options.
   */
  private function getFetcherEndpointFilterTypes() {
    return [
      ProcInterface::FETCHER_FILTER_TYPE_ID => $this->t('ID'),
      ProcInterface::FETCHER_FILTER_TYPE_NAME => $this->t('Name'),
    ];
  }

  /**
   * Return relabelling pattern options.
   */
  private function getRelabelingOptions() {

    $plugin_manager = \Drupal::service('plugin.manager.proc_relabelling');

    // Get list of available plugins.
    $plugins = $plugin_manager->getDefinitions();

    // Populate widget options with plugin labels.
    $options = [];
    $options[] = $this->t('- None -');
    foreach ($plugins as $plugin_id => $plugin_definition) {
      $options[$plugin_id] = $plugin_definition['description'];
    }
    return $options;
  }

  /**
   * {@inheritdoc}
   */
  public static function defaultSettings() {
    return [
      'proc_field_recipients_fetcher_endpoint' => '',
      'proc_field_recipients_fetcher_endpoint_filter' => '',
      'proc_field_recipients_fetcher_endpoint_filter_type' => '',
      'proc_field_recipients_manual_fetcher' => '',
      'proc_field_recipients_to_field' => '',
      'proc_field_recipients_cc_field' => '',
      'proc_field_encrypt_button_label' => '',
      'proc_field_decrypt_button_label' => '',
      'proc_field_mode' => ProcInterface::FIELD_OPS_ONLY_ENCRYPTION,
      'proc_field_input_mode' => ProcInterface::INPUT_MODE_FILE,
      'proc_field_decryption_mode' => '',
      'proc_allow_password_caching' => '',
      'proc_trigger_format' => '',
      'proc_master_field_submit_id' => '',
      'proc_decryption_trigger_fields' => '',
      'proc_relabeling_pattern' => 0,
    ] + parent::defaultSettings();
  }

  /**
   * {@inheritdoc}
   */
  public function settingsForm(array $form, FormStateInterface $form_state) {
    $element = [];
    $element[] = parent::settingsForm($form, $form_state);

    $element['proc_field_recipients_to_field'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Machine name of the field containing the user IDs of recipients'),
      '#description' => $this->t('It must be a user reference field visible in the same form as this field.'),
      '#default_value' => $this->getSetting('proc_field_recipients_to_field') ?? '',
    ];

    $element['proc_field_recipients_cc_field'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Machine name of the field containing user IDs of CC recipients'),
      '#description' => $this->t('It must be a user reference field visible in the same form as this field.'),
      '#default_value' => $this->getSetting('proc_field_recipients_cc_field') ?? '',
    ];

    $element['proc_field_recipients_manual_fetcher'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Define the user IDs of recipients in a CSV list'),
      '#description' => $this->t('Example: 1,2,3'),
      '#default_value' => $this->getSetting('proc_field_recipients_manual_fetcher') ?? '',
      '#disabled' => FALSE,
    ];

    $element['proc_field_recipients_fetcher_endpoint'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Endpoint for fetching user IDs of recipients'),
      '#description' => $this->t('Example: https://example.com/api/get/users/param-a/param-b/param-c?param-d=1&#38;param-e=2'),
      '#default_value' => $this->getSetting('proc_field_recipients_fetcher_endpoint') ?? '',
      '#disabled' => FALSE,
    ];

    $element['proc_field_recipients_fetcher_endpoint_token_tree'] = [
      '#theme' => 'token_tree_link',
      '#token_types' => ['site', 'current-page'],
      '#show_restricted' => TRUE,
    ];

    $element['proc_field_recipients_fetcher_endpoint_filter'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Identifier of form elements to be used as parameters for the endpoint'),
      '#description' => $this->t('Example: edit-field-a,edit-field-b,edit-field-c'),
      '#default_value' => $this->getSetting('proc_field_recipients_fetcher_endpoint_filter') ?? '',
      '#disabled' => FALSE,
    ];

    $element['proc_field_encrypt_button_label'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Label for the encrypt link'),
      '#description' => $this->t('The label for attaching an encrypted file.'),
      '#default_value' => $this->getSetting('proc_field_encrypt_button_label') ?? '',
    ];

    $element['proc_field_decrypt_button_label'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Label for the decrypt link'),
      '#description' => $this->t('The label for decrypt an encrypted file.'),
      '#default_value' => $this->getSetting('proc_field_decrypt_button_label') ?? '',
    ];

    $element['proc_field_recipients_fetcher_endpoint_filter_type'] = [
      '#type' => 'radios',
      '#title' => $this->t('Type of identifier'),
      '#default_value' => $this->getSetting('proc_field_recipients_fetcher_endpoint_filter_type') ?? ProcInterface::FETCHER_FILTER_TYPE_ID,
      '#options' => $this->getFetcherEndpointFilterTypes(),
      '#disabled' => FALSE,
    ];

    $element['proc_field_mode'] = [
      '#type' => 'radios',
      '#title' => $this->t('Modes of operation'),
      '#default_value' => $this->getSetting('proc_field_mode') ?? ProcInterface::FIELD_OPS_ONLY_ENCRYPTION,
      '#options' => $this->getFieldOperationModes(),
      // @todo finish encryption and signature functionality.
      ProcInterface::FIELD_OPS_ENCRYPTION_AND_SIGNATURE => [
        '#disabled' => TRUE,
      ],
    ];

    $element['proc_field_input_mode'] = [
      '#type' => 'radios',
      '#title' => $this->t('Modes of input'),
      '#default_value' => $this->getSetting('proc_field_input_mode') ?? ProcInterface::INPUT_MODE_FILE,
      '#options' => $this->getInputModes(),
      // @todo finish textarea input field.
      ProcInterface::INPUT_MODE_TEXTAREA  => [
        '#disabled' => TRUE,
      ],
      // @todo finish datetime input field.
      ProcInterface::INPUT_MODE_DATE => [
        '#disabled' => TRUE,
      ],
    ];

    $element['proc_field_decryption_mode'] = [
      '#type' => 'radios',
      '#title' => $this->t('Modes of decryption'),
      '#default_value' => $this->getSetting('proc_field_decryption_mode') ?? ProcInterface::DECRYPT_MODE_NEW_PAGE,
      '#options' => $this->getDecryptionModes(),
    ];

    $element['proc_allow_password_caching'] = [
      '#type' => 'radios',
      '#title' => $this->t('Password caching'),
      '#default_value' => $this->getSetting('proc_allow_password_caching') ?? ProcInterface::PASS_CACHE_OPS_DISABLED,
      '#options' => $this->getPasswordCachingOptions(),
      '#description' => $this->t('This configuration does not interfere over Armoured field formatter, where password caching is enabled by default.'),
    ];

    $element['proc_trigger_format'] = [
      '#type' => 'radios',
      '#title' => $this->t('Trigger format'),
      '#default_value' => $this->getSetting('proc_trigger_format') ?? ProcInterface::TRIGGER_FORMAT_BUTTONS,
      '#options' => $this->getTriggerFormats(),
    ];

    $element['proc_master_field_submit_id'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Submit element ID for the master field'),
      '#description' => $this->t('Define the submit element ID of the form where this field is located. Example: edit-submit'),
      '#default_value' => $this->getSetting('proc_master_field_submit_id') ?? '',
      '#disabled' => FALSE,
    ];

    $element['proc_decryption_trigger_fields'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Trigger decryption of another encrypted field'),
      '#default_value' => $this->getSetting('proc_decryption_trigger_fields') ?? '',
    ];

    $element['proc_relabeling_pattern'] = [
      '#type' => 'select',
      '#title' => $this->t('Relabelling pattern'),
      '#default_value' => $this->getSetting('proc_relabeling_pattern') ?? ProcInterface::TRIGGER_FORMAT_BUTTONS,
      '#options' => $this->getRelabelingOptions(),
    ];

    return $element;
  }

  /**
   * {@inheritdoc}
   */
  public function settingsSummary() {
    $summary = parent::settingsSummary();

    $proc_recipients_to_field = $this->getSetting('proc_field_recipients_to_field');
    $summary[] = $this->t('Machine name of recipients To: @proc_field_recipients_to_field', [
      '@proc_field_recipients_to_field' => $proc_recipients_to_field,
    ]);
    $proc_recipients_cc_field = $this->getSetting('proc_field_recipients_cc_field');
    $summary[] = $this->t('Machine name of recipients CC: @proc_field_recipients_cc_field', [
      '@proc_field_recipients_cc_field' => $proc_recipients_cc_field,
    ]);
    $proc_field_recipients_manual_fetcher = $this->getSetting('proc_field_recipients_manual_fetcher');
    $summary[] = $this->t('Default recipients: @proc_field_recipients_manual_fetcher', [
      '@proc_field_recipients_manual_fetcher' => $proc_field_recipients_manual_fetcher,
    ]);
    $proc_field_recipients_fetcher_endpoint = $this->getSetting('proc_field_recipients_fetcher_endpoint');
    $summary[] = $this->t('Recipients fetcher endpoint: @proc_field_recipients_fetcher_endpoint', [
      '@proc_field_recipients_fetcher_endpoint' => $proc_field_recipients_fetcher_endpoint,
    ]);
    $proc_field_recipients_fetcher_endpoint_filter = $this->getSetting('proc_field_recipients_fetcher_endpoint_filter');
    $summary[] = $this->t('Machine name of fetcher filter: @proc_field_recipients_fetcher_endpoint_filter', [
      '@proc_field_recipients_fetcher_endpoint_filter' => ($proc_field_recipients_fetcher_endpoint_filter) ?: '',
    ]);
    $proc_field_recipients_fetcher_endpoint_filter_type = $this->getSetting('proc_field_recipients_fetcher_endpoint_filter_type');
    $summary[] = $this->t('Recipients fetcher type: @proc_field_recipients_fetcher_endpoint_filter_type', [
      '@proc_field_recipients_fetcher_endpoint_filter_type' => ($proc_field_recipients_fetcher_endpoint_filter_type)
        ? $this->getFetcherEndpointFilterTypes()[$proc_field_recipients_fetcher_endpoint_filter_type]
        : '',
    ]);
    $encrypt_button_label = $this->getSetting('proc_field_encrypt_button_label');
    $summary[] = $this->t('Encrypt button label: @encrypt_button_label', [
      '@encrypt_button_label' => ($encrypt_button_label) ?: $this->t('use default'),
    ]);
    $decrypt_button_label = $this->getSetting('proc_field_decrypt_button_label');
    $summary[] = $this->t('Decrypt button label: @decrypt_button_label', [
      '@decrypt_button_label' => ($decrypt_button_label) ?: $this->t('use default'),
    ]);
    $proc_field_mode = $this->getSetting('proc_field_mode');
    $summary[] = $this->t('Operation: @proc_field_mode', [
      '@proc_field_mode' => (isset($proc_field_mode))
        ? $this->getFieldOperationModes()[$proc_field_mode]
        : '',
    ]);
    $proc_field_input_mode = $this->getSetting('proc_field_input_mode');
    $summary[] = $this->t('Input mode: @proc_field_input_mode', [
      '@proc_field_input_mode' => (isset($proc_field_input_mode))
        ? $this->getInputModes()[$proc_field_input_mode]
        : '',
    ]);
    $proc_field_decryption_mode = $this->getSetting('proc_field_decryption_mode');
    $summary[] = $this->t('Decryption mode: @proc_field_decryption_mode', [
      '@proc_field_decryption_mode' => (!empty($proc_field_decryption_mode))
        ? $this->getDecryptionModes()[$proc_field_decryption_mode]
        : '',
    ]);
    $proc_allow_password_caching = $this->getSetting('proc_allow_password_caching');
    $summary[] = $this->t('Password caching: @proc_allow_password_caching', [
      '@proc_allow_password_caching' => (!empty($proc_allow_password_caching))
        ? $this->getPasswordCachingOptions()[$proc_allow_password_caching]
        : '',
    ]);
    $proc_trigger_format = $this->getSetting('proc_trigger_format');
    $summary[] = $this->t('Trigger format: @proc_trigger_format', [
      '@proc_trigger_format' => (!empty($proc_trigger_format))
        ? $this->getTriggerFormats()[$proc_trigger_format]
        : '',
    ]);
    $proc_master_field_submit_id = $this->getSetting('proc_master_field_submit_id');
    $summary[] = $this->t('Submit element ID: @proc_master_field_submit_id', [
      '@proc_master_field_submit_id' => $proc_master_field_submit_id,
    ]);
    $proc_decryption_trigger_fields = $this->getSetting('proc_decryption_trigger_fields');
    $summary[] = $this->t('Trigger other fields on decryption: @proc_decryption_trigger_fields', [
      '@proc_decryption_trigger_fields' => (isset($proc_decryption_trigger_fields))
        ? $this->t('Yes')
        : $this->t('No'),
    ]);
    $proc_relabeling_pattern = $this->getSetting('proc_relabeling_pattern') ?? '0';
    if (!empty($proc_relabeling_pattern)) {
      $summary[] = $this->t('Relabelling pattern: @proc_relabeling_pattern', [
        '@proc_relabeling_pattern' => $this->getRelabelingOptions()[$proc_relabeling_pattern],
      ]);
    }

    return $summary;
  }

  /**
   * {@inheritdoc}
   */
  public function formElement(FieldItemListInterface $items, $delta, array $element, array &$form, FormStateInterface $form_state) {
    // Call the parent formElement method to get the base element.
    $element = parent::formElement($items, $delta, $element, $form, $form_state);
    // Retrieve element forms state value.
    $element_value = $form_state->getValue($items->getName());

    // Extract widget settings into local variables.
    // @todo Move these to class properties.
    $proc_settings = [
      'field_input_mode' => $this->getSetting('proc_field_input_mode') ?? '',
      'to_recipients_field_name' => $this->getSetting('proc_field_recipients_to_field') ?? '',
      'carbon_copy_recipients_field_name' => $this->getSetting('proc_field_recipients_cc_field') ?? '',
      'preset_recipients_csv' => $this->getSetting('proc_field_recipients_manual_fetcher') ?? '',
      'fetcher_endpoint' => $this->getSetting('proc_field_recipients_fetcher_endpoint') ?? '',
      'fetcher_endpoint_filter' => $this->getSetting('proc_field_recipients_fetcher_endpoint_filter') ?? '',
      'fetcher_endpoint_filter_type' => $this->getSetting('proc_field_recipients_fetcher_endpoint_filter_type') ?? '',
      'encrypt_button_label' => $this->getSetting('proc_field_encrypt_button_label') ?? '',
      'decrypt_button_label' => $this->getSetting('proc_field_decrypt_button_label') ?? '',
      'cache_password' => $this->getSetting('proc_allow_password_caching') ?? '0',
      'trigger_format' => $this->getSetting('proc_trigger_format') ?? '0',
      'submit_element_id' => $this->getSetting('proc_master_field_submit_id') ?? '',
      'trigger_fields' => $this->getSetting('proc_decryption_trigger_fields') ?? '0',
      'proc_field_decryption_mode' => $this->getSetting('proc_field_decryption_mode') ?? '',
      'proc_field_name' => $items->getName(),
      'proc_field_mode' => $this->getSetting('proc_field_mode') ?? ProcInterface::FIELD_OPS_DISABLED,
      'proc_relabeling_pattern' => $this->getSetting('proc_relabeling_pattern') ?? ProcInterface::TRIGGER_FORMAT_BUTTONS,
    ];

    // Merge extracted settings with existing element if any.
    $element['#proc-settings'] = array_merge(($element['#proc-settings'] ?? []), $proc_settings);

    if ((int) $element['#proc-settings']['proc_field_mode'] !== ProcInterface::FIELD_OPS_DISABLED) {
      $element['#type'] = 'procfield';
    }

    // Keep form state value after AJAX calls.
    if (!empty($element_value[$delta]['target_id'])) {
      $target_id = $element_value[$delta]['target_id'];
      $proc_entity = \Drupal::entityTypeManager()->getStorage('proc')->load($target_id);
      if ($proc_entity instanceof Proc) {
        // Entity Reference requires " (id)" suffix.
        $element['target_id']['#attributes']['value'] = $proc_entity->label() . ' (' . $target_id . ')';
      }
    }

    $element['#element_validate'][] = [$this, 'validate'];

    $procElement = new ProcField($element, $items->getName(), $items->getEntity()->getEntityTypeId());
    $event = new ProcElementEvent($procElement);
    \Drupal::service('event_dispatcher')->dispatch($event, ProcEvents::PROC_ELEMENT_SETTINGS);

    return $element;
  }

  /**
   * Remove reference if field is empty or keep it if encryption is unchanged.
   */
  public function validate($element, FormStateInterface $form_state) {
    $procField = $element['#proc-settings']['proc_field_name'];

    $default_value = $form_state->getValue($procField)[0]['target_id'] ?? NULL;

    if (empty($default_value)) {
      // This is needed to prevent "This value should be of the correct primitive type." error.
      // See: https://www.drupal.org/project/drupal/issues/2220381
      $form_state->setValueForElement($element, ['target_id' => NULL]);
    }

    if ($default_value === '*******') {
      $form_state->setValueForElement($element, ['target_id' => $element['target_id']['#default_value']->id()]);
    }
    elseif (preg_match('/[*]+ [(]\d+[)]/', (string) $default_value)) {
      $target_id = preg_replace('/[^0-9]+/', '', $default_value);
      $form_state->setValueForElement($element, ['target_id' => $target_id]);
    }

    $values = array_column($form_state->getValue($procField), 'target_id');
    // If this is not a proc text field nor is empty:
    if ($values[0] != '*******' && !empty($values[0]) && ($element['target_id']['#type'] != 'textfield')) {

      if ($this->checkValue($values)) {
        $hostEntity = $element['target_id']['#selection_settings']['entity'] ?? NULL;

        if ($hostEntity) {
          // Load the proc_relabeling_pattern from the field:
          $relabelingPattern = $element['#proc-settings']['proc_relabeling_pattern'];
          $type = \Drupal::service('plugin.manager.proc_relabelling');
          if ($relabelingPattern != '0') {
            foreach ($values as $key => $value) {
              if (is_numeric($value)) {
                $pluginInstance = $type->createInstance($relabelingPattern, []);
                $pluginInstance->relabel([
                  'element' => $element,
                  'form_state' => $form_state,
                  'host_entity' => $hostEntity,
                  'proc_id' => $value,
                  'proc_key' => $key,
                  'proc_field' => $procField,
                ]);
              }
            }
          }
        }
      }
    }
  }

  /**
   * Check if the value is numeric.
   *
   * @param array $values
   *   The values to be checked.
   *
   * @return bool
   */
  private function checkValue($values) {
    if (count($values) > 2) {
      if (array_pop($values) === '') {
        return FALSE;
      }
    }
    foreach ($values as $value) {
      if (!is_numeric($value)) {
        return FALSE;
      }
    }
    return TRUE;
  }

}
